/********************************************************************************
 * Copyright (c) 2020-2023 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/
import {
    AbstractUIExtension,
    codiconCSSClasses,
    EditorContextService,
    GChildElement,
    GLabel,
    GModelRoot,
    hasArgs,
    IActionDispatcher,
    IDiagramStartup,
    MaybePromise,
    TYPES
} from '@eclipse-glsp/client';
import { inject, injectable } from 'inversify';
import { SetPageVisibilityOperation } from '../Actions/SetPageVisibilityOperation';
import { ModelUpdateListener } from '../Handlers/ModelUpdateListener';

@injectable()
export class ActivePagesOverlay extends AbstractUIExtension implements IDiagramStartup {
    @inject(TYPES.IActionDispatcher)
    protected readonly actionDispatcher: IActionDispatcher;

    @inject(EditorContextService)
    protected editorContext: EditorContextService;

    @inject(ModelUpdateListener)
    protected modelUpdateListener: ModelUpdateListener;

    static readonly ID = 'button-overlay';

    id(): string {
        return ActivePagesOverlay.ID;
    }

    containerClass(): string {
        return ActivePagesOverlay.ID;
    }

    protected initializeContents(containerElement: HTMLElement): void {
        this.modelUpdateListener.onModelUpdateComplete(change => {
            this.modelUpdateReload();
        });

        containerElement.appendChild(this.createPageButtonDiv());
    }

    protected createPageButtonDiv(): HTMLElement {
        const buttonContainer = document.createElement('div');
        buttonContainer.id = 'page-button-container';

        const pageNodeList: GChildElement[] = [];
        this.editorContext.modelRoot.children.forEach(child => {
            this.collectPages(pageNodeList, child);
        });

        pageNodeList.forEach(pageNode => {
            let text = '';
            if (pageNode.children[0] instanceof GLabel) {
                text = pageNode.children[0].text;
            }
            let icon = 'chrome-minimize';
            if (hasArgs(pageNode) && pageNode.args.hidden) {
                icon = 'add';
            }

            buttonContainer.appendChild(this.createPageButton('btn_center_diagram', text, icon, pageNode));
        });
        return buttonContainer;
    }

    protected collectPages(pageNodeList: GChildElement[], node: GChildElement): any[] {
        if (node.type === 'pageNode') {
            pageNodeList.push(node);
        }
        node.children.forEach(child => this.collectPages(pageNodeList, child));

        return pageNodeList;
    }

    protected createPageButton(id: string, label: string, codiconId: string, pageNode: GChildElement): HTMLElement {
        const baseDiv = document.getElementById(this.options.baseDiv);
        if (baseDiv) {
            const button = document.createElement('div');
            const insertedDiv = baseDiv.insertBefore(button, baseDiv.firstChild);
            button.id = id + label;
            button.classList.add('overlay-button');
            const icon = this.createIcon(codiconId);
            insertedDiv.appendChild(icon);
            insertedDiv.onclick = () => {
                if (hasArgs(pageNode)) {
                    const isHidden = pageNode.args.hidden;
                    this.actionDispatcher.dispatch(
                        SetPageVisibilityOperation.create(pageNode.id, {
                            hidden: !isHidden
                        })
                    );
                }
            };
            insertedDiv.insertAdjacentText('beforeend', label);
            return button;
        }
        return document.createElement('div');
    }

    protected createIcon(codiconId: string): HTMLElement {
        const icon = document.createElement('i');
        icon.classList.add(...codiconCSSClasses(codiconId), 'overlay-icon');
        return icon;
    }

    postModelInitialization(): MaybePromise<void> {
        this.show(this.editorContext.modelRoot);
    }

    modelUpdateReload(): void {
        document.getElementById('page-button-container')?.replaceWith(this.createPageButtonDiv());
    }

    override show(root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.activeElement = document.activeElement;
        if (!this.containerElement) {
            if (!this.initialize()) {
                return;
            }
        }
        this.onBeforeShow(this.containerElement, root, ...contextElementIds);
        this.setContainerVisible(true);
    }
}
