/********************************************************************************
 * Copyright (c) 2024 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import {
    AbstractUIExtension,
    codiconCSSClasses,
    Disposable,
    DOMHelper,
    EditorContextService,
    GLSPActionDispatcher,
    GLSPMouseTool,
    GModelRoot,
    hasArgs,
    MousePositionTracker,
    TYPES,
    ViewerOptions
} from '@eclipse-glsp/client';
import { inject, injectable } from 'inversify';
import { AddInputSignalOperation } from '../Actions/AddInputSignalOperation';
import { DeleteInputSignalOperation } from '../Actions/DeleteInputSignalOperation';
import { ModelUpdateListener } from '../Handlers/ModelUpdateListener';

@injectable()
export class EditInputSignalMenu extends AbstractUIExtension {
    @inject(TYPES.IActionDispatcher)
    protected actionDispatcher: GLSPActionDispatcher;

    @inject(EditorContextService)
    protected editorContextService: EditorContextService;

    @inject(MousePositionTracker)
    protected mousePositionTracker: MousePositionTracker;

    @inject(TYPES.ViewerOptions)
    protected viewerOptions: ViewerOptions;

    @inject(TYPES.DOMHelper)
    protected domHelper: DOMHelper;

    @inject(GLSPMouseTool) protected mouseTool: GLSPMouseTool;

    @inject(ModelUpdateListener)
    protected modelUpdateListener: ModelUpdateListener;

    protected waitForUpdate: Disposable;
    protected waitForUpdateIsDisposed = true;

    static readonly ID = 'edit-signal-menu';

    override id(): string {
        return EditInputSignalMenu.ID;
    }
    override containerClass(): string {
        return 'command-palette';
    }
    protected override initializeContents(containerElement: HTMLElement): void {
        containerElement.appendChild(this.createWidget());
    }

    protected createWidget(): HTMLElement {
        const editingElement = document.createElement('div');
        editingElement.id = 'input-signal-editing-element';
        editingElement.classList.add('edit-box');
        const header = document.createElement('div');

        header.appendChild(document.createTextNode('Edit Input Signals'));

        const closeIcon = this.createIcon('chrome-close', 'Close');
        closeIcon.style.margin = '0px 0px 0px 0px';

        closeIcon.onclick = () => {
            this.hide();
        };

        header.appendChild(closeIcon);
        header.classList.add('edit-box-header');
        editingElement.appendChild(header);

        editingElement.appendChild(this.createInputSection());
        const sepLine = document.createElement('hr');
        sepLine.style.margin = '4px 0px 0px';

        editingElement.appendChild(sepLine);
        editingElement.appendChild(this.createDeletionSection());
        return editingElement;
    }

    protected createInputSection(): HTMLElement {
        const inputSection = document.createElement('div');
        const titleDiv = document.createElement('div');
        const title = document.createTextNode('Add Signal:');
        titleDiv.appendChild(title);
        titleDiv.classList.add('edit-box-text');

        inputSection.appendChild(titleDiv);
        const inputField = document.createElement('input');
        inputField.id = 'add-signal-input-field';
        inputField.style.width = '75%';
        inputField.style.border = 'solid';
        inputField.style.borderWidth = '1px';
        inputField.style.borderColor = '#DDDDDD';
        inputSection.appendChild(inputField);
        const addButton = document.createElement('button');
        addButton.appendChild(document.createTextNode('Add'));
        addButton.style.color = '#333333';
        addButton.onclick = async () => {
            if (this.waitForUpdateIsDisposed) {
                const signalId = inputField.value;
                if (!signalId) {
                    return;
                }

                await this.actionDispatcher.dispatch(AddInputSignalOperation.create(signalId));
                this.waitForUpdate = this.modelUpdateListener.onModelUpdateComplete(change => {
                    this.modelUpdateReload();
                });
                this.waitForUpdateIsDisposed = false;
            }
        };
        inputSection.appendChild(addButton);
        return inputSection;
    }

    protected createDeletionSection(): HTMLElement {
        const deletionSection = document.createElement('div');
        const titleDiv = document.createElement('div');
        const title = document.createTextNode('Input Signals:');
        titleDiv.appendChild(title);
        titleDiv.classList.add('edit-box-text');

        deletionSection.appendChild(titleDiv);
        const root = this.editorContextService.modelRoot;
        const graphArgs = hasArgs(root) ? root.args : {};
        // eslint-disable-next-line guard-for-in
        for (const signalId in graphArgs) {
            deletionSection.appendChild(this.createDeletionElement(signalId));
        }

        return deletionSection;
    }

    protected createDeletionElement(signalId: string): HTMLElement {
        const deletionElement = document.createElement('div');
        deletionElement.classList.add('delete-field');
        const icon = this.createIcon('chrome-close', 'Delete');

        icon.onclick = async () => {
            if (this.waitForUpdateIsDisposed) {
                await this.actionDispatcher.dispatch(DeleteInputSignalOperation.create(signalId));
                this.waitForUpdate = this.modelUpdateListener.onModelUpdateComplete(change => {
                    this.modelUpdateReload();
                });
                this.waitForUpdateIsDisposed = false;
            }
        };
        deletionElement.appendChild(icon);
        deletionElement.insertAdjacentText('beforeend', signalId);
        return deletionElement;
    }

    protected createIcon(codiconId: string, hoverText: string): HTMLElement {
        const icon = document.createElement('i');
        icon.classList.add(...codiconCSSClasses(codiconId), 'overlay-icon', 'delete-icon');
        icon.setAttribute('title', hoverText);
        return icon;
    }

    override show(root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        super.show(root, ...contextElementIds);
    }

    protected override onBeforeShow(containerElement: HTMLElement, root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.setPosition(containerElement);
    }

    protected setPosition(containerElement: HTMLElement): void {
        this.containerElement.style.position = 'absolute';
        const x = 25;
        const y = 100;

        containerElement.style.left = `${x}px`;
        containerElement.style.top = `${y}px`;
        containerElement.style.width = '200px';
    }

    override hide(): void {
        super.hide();
    }

    modelUpdateReload(): void {
        document.getElementById('input-signal-editing-element')?.replaceWith(this.createWidget());
        this.waitForUpdate?.dispose();
        this.waitForUpdateIsDisposed = true;
    }
}
