/********************************************************************************
 * Copyright (c) 2024 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import {
    AbstractUIExtension,
    AutoCompleteWidget,
    DOMHelper,
    EditorContextService,
    getAbsoluteClientBounds,
    GLSPActionDispatcher,
    GModelRoot,
    GNode,
    LabeledAction,
    ModelIndexImpl,
    MousePositionTracker,
    RequestEditValidationAction,
    SetEditValidationResultAction,
    TYPES,
    ValidationDecorator,
    ValidationStatus,
    ViewerOptions
} from '@eclipse-glsp/client';
import { inject, injectable } from 'inversify';
import { EditSignalClauseOperation } from '../Actions/EditSignalClauseOperation';
import { isTransition } from '../model';

@injectable()
export class SignalClauseMenu extends AbstractUIExtension {
    @inject(TYPES.IActionDispatcher)
    protected actionDispatcher: GLSPActionDispatcher;

    @inject(EditorContextService)
    protected editorContextService: EditorContextService;

    @inject(MousePositionTracker)
    protected mousePositionTracker: MousePositionTracker;

    @inject(TYPES.ViewerOptions)
    protected viewerOptions: ViewerOptions;

    @inject(TYPES.DOMHelper)
    protected domHelper: DOMHelper;

    static readonly ID = 'signal-clause-menu';

    protected transition: GNode;
    protected autoSuggestion: AutoCompleteWidget;
    readonly autoSuggestionSettings = {
        noSuggestionsMessage: '',
        suggestionsClass: 'command-palette-suggestions',
        debounceWaitMs: 50,
        showOnFocus: true
    };

    override id(): string {
        return SignalClauseMenu.ID;
    }
    override containerClass(): string {
        return 'command-palette';
    }
    protected override initializeContents(containerElement: HTMLElement): void {
        this.autoSuggestion = new AutoCompleteWidget(
            this.autoSuggestionSettings,
            { provideSuggestions: input => this.retrieveSuggestions(input) },
            { executeFromSuggestion: input => {} },
            () => this.hide(),
            this.logger
        );
        this.autoSuggestion.configureValidation(
            { validate: input => this.validateInput(input) },
            new ValidationDecorator(containerElement)
        );
        this.autoSuggestion.configureTextSubmitHandler({
            executeFromTextOnlyInput: (input: string) => this.executeFromTextOnlyInput(input)
        });
        this.autoSuggestion.initialize(containerElement);
    }

    protected async validateInput(input: string): Promise<ValidationStatus> {
        const response = await this.actionDispatcher.request(
            RequestEditValidationAction.create({ contextId: SignalClauseMenu.ID, modelElementId: this.transition.id, text: input })
        );
        if (SetEditValidationResultAction.is(response)) {
            return response.status;
        }
        return Promise.reject();
    }

    override show(root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        super.show(root, ...contextElementIds);
        this.autoSuggestion.open(root);
    }

    protected override onBeforeShow(containerElement: HTMLElement, root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.transition = getTransition(contextElementIds, root.index)[0];
        this.autoSuggestion.inputField.value = '';
        this.setPosition(containerElement);
    }

    protected setPosition(containerElement: HTMLElement): void {
        let x = 0;
        let y = 0;

        if (this.transition) {
            const bounds = getAbsoluteClientBounds(this.transition, this.domHelper, this.viewerOptions);
            x = bounds.x + 5;
            y = bounds.y + 5;
        }
        containerElement.style.left = `${x}px`;
        containerElement.style.top = `${y}px`;
        containerElement.style.width = '200px';
    }

    executeFromTextOnlyInput(input: string): void {
        this.actionDispatcher.dispatch(EditSignalClauseOperation.create(this.transition.id, input));
    }

    protected async retrieveSuggestions(input: string): Promise<LabeledAction[]> {
        return [];
    }

    override hide(): void {
        this.autoSuggestion.dispose();
        super.hide();
    }
}

function getTransition(ids: string[], index: ModelIndexImpl): GNode[] {
    return ids.map(id => index.getById(id)).filter(element => element && isTransition(element)) as GNode[];
}
