/********************************************************************************
 * Copyright (c) 2024 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied: GNU General Public License, version 2
 * with the GNU Classpath Exception which is available at
 * https://www.gnu.org/software/classpath/license.html.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
 ********************************************************************************/

import {
    AbstractUIExtension,
    codiconCSSClasses,
    Disposable,
    DOMHelper,
    EditorContextService,
    GLSPActionDispatcher,
    GLSPMouseTool,
    GModelRoot,
    MousePositionTracker,
    TYPES,
    ViewerOptions
} from '@eclipse-glsp/client';
import { inject, injectable } from 'inversify';
import { RunTinaAction } from '../Actions/RunTinaAction';
import { TinaResponseAction } from '../Actions/TinaResponseAction';
import { ModelUpdateListener } from '../Handlers/ModelUpdateListener';

@injectable()
export class TinaMenu extends AbstractUIExtension {
    @inject(TYPES.IActionDispatcher)
    protected actionDispatcher: GLSPActionDispatcher;

    @inject(EditorContextService)
    protected editorContextService: EditorContextService;

    @inject(MousePositionTracker)
    protected mousePositionTracker: MousePositionTracker;

    @inject(TYPES.ViewerOptions)
    protected viewerOptions: ViewerOptions;

    @inject(TYPES.DOMHelper)
    protected domHelper: DOMHelper;

    @inject(GLSPMouseTool) protected mouseTool: GLSPMouseTool;

    @inject(ModelUpdateListener)
    protected modelUpdateListener: ModelUpdateListener;

    protected waitForUpdate: Disposable;
    protected waitForUpdateIsDisposed = true;

    static readonly ID = 'tina-menu';

    override id(): string {
        return TinaMenu.ID;
    }
    override containerClass(): string {
        return 'command-palette';
    }
    protected override initializeContents(containerElement: HTMLElement): void {
        containerElement.appendChild(this.createWidget());
    }

    protected createWidget(): HTMLElement {
        const editingElement = document.createElement('div');
        editingElement.id = 'tina-element';
        editingElement.classList.add('edit-box');
        const header = document.createElement('div');

        header.appendChild(document.createTextNode('Run Tina'));

        const closeIcon = this.createIcon('chrome-close', 'Close');
        closeIcon.style.margin = '0px 0px 0px 0px';

        closeIcon.onclick = () => {
            this.hide();
        };

        header.appendChild(closeIcon);
        header.classList.add('edit-box-header');
        editingElement.appendChild(header);

        editingElement.appendChild(this.createInputSection());
        const sepLine = document.createElement('hr');
        sepLine.style.margin = '4px 0px 0px';

        editingElement.appendChild(sepLine);

        editingElement.appendChild(this.createResponseSection(''));

        return editingElement;
    }

    protected createInputSection(): HTMLElement {
        const inputSection = document.createElement('div');

        const inputField = document.createElement('input');
        inputField.id = 'add-signal-input-field';
        inputField.style.width = '85%';
        inputField.style.border = 'solid';
        inputField.style.borderWidth = '1px';
        inputField.style.borderColor = '#DDDDDD';
        inputSection.appendChild(inputField);

        const addButton = document.createElement('button');
        addButton.appendChild(document.createTextNode('Run'));
        addButton.style.color = '#333333';
        addButton.onclick = async () => {
            const tinaCommand = inputField.value;
            if (!tinaCommand) {
                return;
            }

            const response = await this.actionDispatcher.request(RunTinaAction.create(tinaCommand));
            inputField.value = '';

            if (TinaResponseAction.is(response)) {
                this.displayResponse(response.tinaResponse);
            }
            return Promise.reject();
        };
        inputSection.appendChild(addButton);
        return inputSection;
    }

    protected createResponseSection(response: string): HTMLElement {
        const responseSection = document.createElement('div');
        responseSection.id = 'response-section-element';
        if (response) {
            responseSection.appendChild(document.createTextNode(response));
            responseSection.style.padding = '0.4em';
        }
        return responseSection;
    }

    protected createIcon(codiconId: string, hoverText: string): HTMLElement {
        const icon = document.createElement('i');
        icon.classList.add(...codiconCSSClasses(codiconId), 'overlay-icon', 'delete-icon');
        icon.setAttribute('title', hoverText);
        return icon;
    }

    override show(root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        super.show(root, ...contextElementIds);
    }

    protected override onBeforeShow(containerElement: HTMLElement, root: Readonly<GModelRoot>, ...contextElementIds: string[]): void {
        this.setPosition(containerElement);
    }

    protected setPosition(containerElement: HTMLElement): void {
        this.containerElement.style.position = 'absolute';
        const x = 250;
        const y = 10;

        containerElement.style.left = `${x}px`;
        containerElement.style.top = `${y}px`;
        containerElement.style.width = '400px';
    }

    override hide(): void {
        document.getElementById('response-section-element')?.replaceWith(this.createResponseSection(''));
        super.hide();
    }

    displayResponse(response: string): void {
        document.getElementById('response-section-element')?.replaceWith(this.createResponseSection(response));
    }
}
