/********************************************************************************
 * Copyright (c) 2022 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * https://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied:
 * -- GNU General Public License, version 2 with the GNU Classpath Exception
 * which is available at https://www.gnu.org/software/classpath/license.html
 * -- MIT License which is available at https://opensource.org/license/mit.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0 OR MIT
 ********************************************************************************/
package org.eclipse.glsp.example.javaemf;

import org.eclipse.glsp.example.javaemf.MQTT.MQTTClient;
import org.eclipse.glsp.example.javaemf.Simulation.FireTransitionOperationHandler;
import org.eclipse.glsp.example.javaemf.Simulation.ResetMarkingsOperationHandler;
import org.eclipse.glsp.example.javaemf.TinaMenu.PNMLFlattener;
import org.eclipse.glsp.example.javaemf.TinaMenu.RunTinaActionHandler;
import org.eclipse.glsp.example.javaemf.TinaMenu.TinaMenuContextActionProvider;
import org.eclipse.glsp.example.javaemf.handler.*;
import org.eclipse.glsp.example.javaemf.handler.DiNeROS.*;
import org.eclipse.glsp.example.javaemf.handler.DiNeROS.ElementCreation.*;
import org.eclipse.glsp.example.javaemf.handler.EdgeAndRefCreation.*;
import org.eclipse.glsp.example.javaemf.handler.NodeCreation.*;
import org.eclipse.glsp.example.javaemf.Simulation.FireTransitionExecutor;
import org.eclipse.glsp.example.javaemf.validators.*;
import org.eclipse.glsp.example.javaemf.model.*;
import org.eclipse.glsp.example.javaemf.palette.TaskListToolPaletteItemProvider;
import org.eclipse.glsp.server.actions.ActionDispatcher;
import org.eclipse.glsp.server.actions.ActionHandler;
import org.eclipse.glsp.server.di.MultiBinding;
import org.eclipse.glsp.server.diagram.DiagramConfiguration;
import org.eclipse.glsp.server.emf.EMFIdGenerator;
import org.eclipse.glsp.server.emf.EMFSourceModelStorage;
import org.eclipse.glsp.server.emf.idgen.AttributeIdGenerator;
import org.eclipse.glsp.server.emf.notation.EMFChangeBoundsOperationHandler;
import org.eclipse.glsp.server.emf.notation.EMFNotationDiagramModule;
import org.eclipse.glsp.server.emf.notation.EMFNotationModelState;
import org.eclipse.glsp.server.features.contextactions.ContextActionsProvider;
import org.eclipse.glsp.server.features.core.model.GModelFactory;
import org.eclipse.glsp.server.features.directediting.ContextEditValidator;
import org.eclipse.glsp.server.features.directediting.LabelEditValidator;
import org.eclipse.glsp.server.features.toolpalette.ToolPaletteItemProvider;
import org.eclipse.glsp.server.features.typehints.EdgeCreationChecker;
import org.eclipse.glsp.server.internal.actions.DefaultActionDispatcher;
import org.eclipse.glsp.server.operations.OperationActionHandler;
import org.eclipse.glsp.server.operations.OperationHandler;

public class TaskListDiagramModule extends EMFNotationDiagramModule {

   @Override
   protected Class<? extends DiagramConfiguration> bindDiagramConfiguration() {
      // define what operations are allowed with our elements
      return TaskListDiagramConfiguration.class;
   }

   @Override
   protected Class<? extends EMFSourceModelStorage> bindSourceModelStorage() {
      // ensure our custom package is registered when loading our models
      return TaskListSourceModelStorage.class;
   }

   @Override
   public Class<? extends GModelFactory> bindGModelFactory() {
      // custom factory to convert tasks into nodes
      return TaskListGModelFactory.class;
   }

   @Override
   protected Class<? extends EMFIdGenerator> bindEMFIdGenerator() {
      // all our elements inherit from Identifiable and have an ID attribute set
      return AttributeIdGenerator.class;
   }

   @Override
   protected Class<? extends ToolPaletteItemProvider> bindToolPaletteItemProvider() {
      return TaskListToolPaletteItemProvider.class;
   }

   @Override
   protected void configureOperationHandlers(final MultiBinding<OperationHandler<?>> binding) {
      super.configureOperationHandlers(binding);
      binding.add(CreatePlaceHandler.class);
      binding.add(CreateTransitionHandler.class);
      binding.add(CreateArcHandler.class);
      binding.add(CreatePageHandler.class);
      binding.add(CreateReferencePlaceHandler.class);
      binding.add(CreateReferenceTransitionHandler.class);
      binding.add(CreateReferenceHandler.class);
      binding.add(SetPageVisibilityHandler.class);
      binding.add(CreateInhibitorArcHandler.class);
      binding.add(CreateReadArcHandler.class);
      binding.add(LabelEditOperationHandler.class);
      binding.rebind(EMFChangeBoundsOperationHandler.class, PnmlChangeBoundsOperationHandler.class);
      binding.add(ChangeRoutingPointsOperationHandler.class);
      binding.add(DeleteOperationHandler.class);
      binding.add(FireTransitionOperationHandler.class);
      binding.add(ResetMarkingsOperationHandler.class);
      binding.add(CreateTopicTransitionHandler.class);
      binding.add(CreateServiceTransitionHandler.class);
      binding.add(CreateTopicLinkHandler.class);
      binding.add(CreateServiceChannelHandler.class);
      binding.add(CreateServiceChannelLinkHandler.class);
      binding.add(CreateServiceServerLinkHandler.class);
      binding.add(CreateServerPageHandler.class);
      binding.add(EditSignalClauseHandler.class);
      binding.add(DeleteInputSignalHandler.class);
      binding.add(AddInputSignalHandler.class);
      binding.add(SetInputSignalHandler.class);
      binding.add(AddDiNeROSTokenHandler.class);
      binding.add(DeleteDiNeROSTokenHandler.class);
      binding.add(SetDiNeROSTokenHandler.class);

      binding.add(CreateNodePageNodeHandler.class);
   }

   @Override
   protected Class<? extends ActionDispatcher> bindActionDispatcher() {
      return MyActionDispatcher.class;
   }
   /**/
   @Override
   protected void configureActionHandlers(final MultiBinding<ActionHandler> binding) {
      super.configureActionHandlers(binding);
      binding.rebind(OperationActionHandler.class, ExtendedOperationActionHandler.class);
      binding.add(RunTinaActionHandler.class);
      binding.rebind(DefaultActionDispatcher.class, MyActionDispatcher.class);

   }

   @Override
   protected Class<? extends LabelEditValidator> bindLabelEditValidator() {
      return PnmlLabelEditValidator.class;
   }

   @Override
   protected Class<? extends EMFNotationModelState> bindGModelState() {
      return ExtendedModelState.class;
   }

   @Override
   protected void configureContextActionsProviders(final MultiBinding<ContextActionsProvider> binding) {
      super.configureContextActionsProviders(binding);
      binding.add(TinaMenuContextActionProvider.class);
   }

   @Override
   protected void configureContextEditValidators(final MultiBinding<ContextEditValidator> binding) {
      super.configureContextEditValidators(binding);
      binding.add(SignalClauseMenuValidator.class);
   }

   @Override
   public String getDiagramType() { return "tasklist-diagram"; }

   @Override
   protected Class<? extends EdgeCreationChecker> bindEdgeCreationChecker() {
      return PNMLEdgeCreationChecker.class;
   }

   protected Class<? extends FireTransitionExecutor> bindFireTransitionExecutor() {
      return FireTransitionExecutor.class;
   }

   protected Class<? extends ModelAccessUtil> bindModelAccessUtil() {
      return ModelAccessUtil.class;
   }

   protected Class<? extends DiNeROSModelFactory> bindDiNeROSFactory() {
      return DiNeROSModelFactory.class;
   }

   protected Class<? extends MQTTClient> bindMQTTClient() {
      return MQTTClient.class;
   }

   protected Class<? extends InputSignalValidator> bindInputSignalValidator() {
      return InputSignalValidator.class;
   }

   protected Class<? extends PNMLFlattener> bindPNMLFlattener() {
      return PNMLFlattener.class;
   }

   protected Class<? extends CyclicReferencesValidator> bindCyclicReferencesValidator() {
      return CyclicReferencesValidator.class;
   }


}
