package org.eclipse.glsp.example.javaemf.TinaMenu;

import jakarta.inject.Inject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.glsp.example.javaemf.model.ExtendedModelState;
import org.eclipse.glsp.example.javaemf.model.ModelAccessUtil;
import org.pnml.emf.ptnet.*;

import java.util.*;

public class PNMLFlattener {
    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected ModelAccessUtil modelAccessUtil;

    protected ArrayList<Arc> inhibitorArcs = new ArrayList<>();

    public pnml flatten() {

        pnml pnmlModel = (pnml) EcoreUtil.copy(modelState.getSemanticModel());

        Page mainPage = flattenToPage(pnmlModel);
        removeReferences(mainPage);
        extractAndRemoveSpecialArcs(mainPage);

        pnmlModel.getNet().get(0).getPage().clear();
        pnmlModel.getNet().get(0).getPage().add(mainPage);

        return pnmlModel;
    }

    protected Page flattenToPage(pnml pnmlModel){
        Page mainPage = PtnetFactory.eINSTANCE.createPage();
        mainPage.setId(UUID.randomUUID().toString());

        pnmlModel.getNet().get(0).getPage().forEach(page -> {
            extractPageContents(page, mainPage);
            page.getPage().forEach(subpage ->
                extractPageContents(subpage, mainPage));
        });

        return mainPage;
    }

    protected void extractPageContents(Page page, Page mainPage){
        mainPage.getPlace().addAll(page.getPlace());
        mainPage.getTransition().addAll(
            page.getTransition().stream().filter(transition -> // removes DiNeROS elements
                !modelAccessUtil.isTopicTransition(transition)
                    && !modelAccessUtil.isServiceTransition(transition)
            ).toList());
        mainPage.getReferencePlace().addAll(page.getReferencePlace());
        mainPage.getReferenceTransition().addAll(page.getReferenceTransition());
        mainPage.getArc().addAll(page.getArc());
    }

    protected Page removeReferences(Page mainPage){
        HashSet<Arc> arcsToRemove = new HashSet<>();
        mainPage.getReferencePlace().forEach(refPlace -> {
            Place referencedPlace = resolvePlaceRef(refPlace);
            Arc[] inArcs = new Arc[refPlace.getInArcs().size()];
            refPlace.getInArcs().toArray(inArcs);

            for (Arc arc : inArcs) {
                if (referencedPlace == null){
                    arcsToRemove.add(arc);
                } else {
                    arc.setTarget(referencedPlace);
                }
            }


            Arc[] outArcs = new Arc[refPlace.getOutArcs().size()];
            refPlace.getOutArcs().toArray(outArcs);

            for (Arc arc : outArcs) {
                if (referencedPlace == null){
                    arcsToRemove.add(arc);
                } else {
                    arc.setSource(referencedPlace);
                }
            }
        });

        mainPage.getReferencePlace().clear();
        mainPage.getReferenceTransition().forEach(refTransition -> {
            Transition referencedTransition = resolveTransitionRef(refTransition);
            Arc[] inArcs = new Arc[refTransition.getInArcs().size()];
            refTransition.getInArcs().toArray(inArcs);
            for (Arc arc : inArcs) {
                if (referencedTransition == null){
                    arcsToRemove.add(arc);
                } else {
                    arc.setTarget(referencedTransition);
                }
            }
            Arc[] outArcs = new Arc[refTransition.getOutArcs().size()];
            refTransition.getOutArcs().toArray(outArcs);

            for (Arc arc : outArcs) {
                if (referencedTransition == null){
                    arcsToRemove.add(arc);
                } else {
                    arc.setSource(referencedTransition);
                }
            }
        });
        mainPage.getReferenceTransition().clear();

        mainPage.getPlace().forEach(place -> {
            place.getReferencingPlaces().clear();
        });

        mainPage.getTransition().forEach(transition -> {
            transition.getReferencingTransitions().clear();
        });

        arcsToRemove.forEach(arc -> {
            arc.getSource().getOutArcs().remove(arc);
            arc.getTarget().getInArcs().remove(arc);
            mainPage.getArc().remove(arc);
        }
        );

        return mainPage;
    }


    protected Place resolvePlaceRef(RefPlace refPlace){
        PlaceNode place = refPlace.getRef();
        while (place instanceof RefPlace){
            place = ((RefPlace) place).getRef();
        }
        if (place instanceof Place){
            return (Place) place;
        } else {
            return null;
        }
    }

    protected Transition resolveTransitionRef(RefTransition refTransition){
        TransitionNode transition = refTransition.getRef();
        while (transition instanceof RefTransition){
            transition = ((RefTransition) transition).getRef();
        }
        if (transition instanceof Transition){
            return (Transition) transition;
        } else {
            return null;
        }
    }

    protected void extractAndRemoveSpecialArcs(Page mainPage){
        ArrayList<Arc> inhibitorArcsTemp = new ArrayList<>();
        // copy so the inhibitor arcs are not affected by removing them from the mainPage (source and target would be removed)
        ArrayList<Arc> inhibitorArcsCopy = new ArrayList<>();
        ArrayList<Arc> readArcs = new ArrayList<>();

        mainPage.getArc().forEach(arc -> {
            if(modelAccessUtil.isInhibitorArc(arc)){
                inhibitorArcsTemp.add(arc);
            }
            if(modelAccessUtil.isReadArc(arc)){
                readArcs.add(arc);
            }
        });

        inhibitorArcsTemp.forEach(arc -> {
            // removes inhibitor arcs from net; they can later be readded to a .net file
            Arc arcCopy = EcoreUtil.copy(arc);
            arcCopy.setSource(EcoreUtil.copy(arc.getSource()));
            arcCopy.setTarget(EcoreUtil.copy(arc.getTarget()));
            inhibitorArcsCopy.add(arcCopy);

            //arc.getSource().getOutArcs().remove(arc);
            //arc.getTarget().getInArcs().remove(arc);
            //mainPage.getArc().remove(arc);
        });

        this.inhibitorArcs = inhibitorArcsCopy;

        readArcs.forEach(arc -> {
            // creates an arc in the other direction
            // is in most cases semantically equivalent to a read arc
            Arc reverseArc = PtnetFactory.eINSTANCE.createArc();
            reverseArc.setSource(arc.getTarget());
            reverseArc.setTarget(arc.getSource());
            reverseArc.setInscription(arc.getInscription());
            reverseArc.setId(arc.getId() + "_reverse");
            mainPage.getArc().add(reverseArc);
        });
    }

    public ArrayList<Arc> getInhibitorArcs(){
        return inhibitorArcs;
    }

}
