package org.eclipse.glsp.example.javaemf.handler;

import com.google.inject.Inject;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.edit.command.RemoveCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.emf.edit.domain.EditingDomain;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.example.javaemf.model.ModelAccessUtil;
import org.eclipse.glsp.graph.GEdge;
import org.eclipse.glsp.graph.GModelElement;
import org.eclipse.glsp.server.emf.EMFOperationHandler;
import org.eclipse.glsp.server.emf.notation.EMFNotationModelIndex;
import org.eclipse.glsp.server.emf.notation.EMFNotationModelState;
import org.eclipse.glsp.server.operations.DeleteOperation;
import org.pnml.emf.ptnet.*;

import java.util.*;

public class DeleteOperationHandler extends EMFOperationHandler<DeleteOperation> {

   @Inject
   protected EMFNotationModelState modelState;

   @Inject
   protected ModelAccessUtil modelAccessUtil;

   protected Set<String> alreadyDeletedArcs;

   @Override
   public Optional<Command> createCommand(final DeleteOperation operation) {
      List<String> elementIds = operation.getElementIds();
      if (elementIds == null || elementIds.isEmpty()) {
         System.out.println("Elements to delete are not specified");
         return Optional.empty();
      }

      List<Command> commands = createDeleteCommands(elementIds);
      return commands.isEmpty() ? Optional.empty() : Optional.of(new CompoundCommand(commands));
   }

   private List<Command> createDeleteCommands(final List<String> elementIds) {
      EMFNotationModelIndex index = modelState.getIndex();
      alreadyDeletedArcs = new HashSet<>();

      List<Command> commands = new ArrayList<>();
      for (String elementId : elementIds) {
         Optional<EObject> semanticElement = index.get(elementId).flatMap(index::getEObject);
         if (semanticElement.isPresent()){
            semanticElement.map(this::createDependentRemoveCommands).ifPresent(commands::addAll);
         } else {
            removeUnindexedElement(elementId).ifPresent(commands::add);
         }


      }
      return commands;
   }

   protected List<Command> createDependentRemoveCommands(EObject element){
      EditingDomain editingDomain = modelState.getEditingDomain();
      List<Command> commands = new ArrayList<>();


      if (element instanceof Node node) {
         // remove node specifics
            // edges
         node.getOutArcs().forEach(arc -> commands.addAll(createRemoveArcCommand(arc)));
         node.getInArcs().forEach(arc -> commands.addAll(createRemoveArcCommand(arc)));

         if (element instanceof PlaceNode placeNode){
            // remove place refs
            //placeNode.getReferencingPlaces().forEach(refPlace -> commands.add(
            //    SetCommand.create(editingDomain, refPlace, PtnetPackage.Literals.REF_PLACE__REF, null)
            //));

            if (element instanceof RefPlace refPlace && refPlace.getRef() != null) {
                // remove form referencingPlaces
                commands.add(
                   RemoveCommand.create(modelState.getEditingDomain(), refPlace.getRef(),
                       PtnetPackage.Literals.PLACE_NODE__REFERENCING_PLACES, refPlace)
               );
            }
         } else if (element instanceof TransitionNode transitionNode) {
            // remove transition refs
            //transitionNode.getReferencingTransitions().forEach(refTransition -> commands.add(
            //    SetCommand.create(editingDomain, refTransition, PtnetPackage.Literals.REF_TRANSITION__REF, null)
            //));

            if (element instanceof RefTransition refTransition && refTransition.getRef() != null) {
               // remove form referencingTransitions
               commands.add(
                   RemoveCommand.create(modelState.getEditingDomain(), refTransition.getRef(),
                       PtnetPackage.Literals.TRANSITION_NODE__REFERENCING_TRANSITIONS, refTransition)
               );
            }
         }
      } else if (element instanceof Page page) {
         // remove for all children
         // needs to be done because the children of the page might be referenced by
         // other nodes that not on the page
         page.getArc().forEach(arc -> commands.addAll(createDependentRemoveCommands(arc)));
         page.getPlace().forEach(place -> commands.addAll(createDependentRemoveCommands(place)));
         page.getTransition().forEach(transition -> commands.addAll(createDependentRemoveCommands(transition)));
         page.getReferencePlace().forEach(refPlace -> commands.addAll(createDependentRemoveCommands(refPlace)));
         page.getReferenceTransition().forEach(refTransition -> commands.addAll(createDependentRemoveCommands(refTransition)));
         page.getPage().forEach(subPage -> commands.addAll(createDependentRemoveCommands(subPage)));

      } else if (element instanceof Arc arc) {
         // remove out and in arcs
         commands.addAll(createRemoveArcCommand(arc));
         return commands;
      }

      commands.add(
          RemoveCommand.create(editingDomain, element.eContainer(), element.eContainingFeature(), element)
      );

      return commands;
   }

   protected List<Command> createRemoveArcCommand(Arc arc){
      EditingDomain editingDomain = modelState.getEditingDomain();
      List<Command> commands = new ArrayList<>();

      // prevents duplicate remove commands for arcs since they can also be removed as a consequence
      // of removing their source/target node
      if (alreadyDeletedArcs.contains(arc.getId())) {
         return commands;
      }

      alreadyDeletedArcs.add(arc.getId());

      commands.add(
          RemoveCommand.create(editingDomain, arc.getSource(),
              PtnetPackage.Literals.NODE__OUT_ARCS, arc)
      );
      commands.add(
          RemoveCommand.create(editingDomain, arc.getTarget(),
              PtnetPackage.Literals.NODE__IN_ARCS, arc)
      );
      commands.add(
          RemoveCommand.create(editingDomain, arc.eContainer(), arc.eContainingFeature(), arc)
      );

      return commands;
   }

   // needed for some DiNeROS elements that are not in the index since they are references of other elements
   protected Optional<Command> removeUnindexedElement(String elementId){
      EditingDomain editingDomain = modelState.getEditingDomain();
      EMFNotationModelIndex index = modelState.getIndex();
      if(index.get(elementId).isEmpty()){
         return Optional.empty();
      }
      GModelElement gElement = index.get(elementId).get();
      if (PNMLModelTypes.TOPIC_LINK.equals(gElement.getType())){
         GEdge edge = (GEdge) gElement;
         if (PNMLModelTypes.TOPIC_TRANSITION.equals(edge.getSource().getType())){
            Transition topicTransition = (Transition) index.getEObject(edge.getSourceId()).orElseThrow();
            Subscribers subs =  modelAccessUtil.getDiNeROSTransitionToolSpecific(topicTransition, PNMLModelTypes.TOPIC_TRANSITION)
                .map(ToolInfoTransition::getSubscribers).orElseThrow();
            Subscriber sub = subs.getSubscriber().stream().filter(subscriber -> Objects.equals(subscriber.getId(), edge.getTargetId()))
                        .findFirst().orElseThrow();
            return Optional.of(RemoveCommand.create(editingDomain, subs, PtnetPackage.Literals.SUBSCRIBERS__SUBSCRIBER, sub));
         } else {
            Transition topicTransition = (Transition) index.getEObject(edge.getTargetId()).orElseThrow();
            Publishers pubs =  modelAccessUtil.getDiNeROSTransitionToolSpecific(topicTransition, PNMLModelTypes.TOPIC_TRANSITION)
                .map(ToolInfoTransition::getPublishers).orElseThrow();
            Publisher sub = pubs.getPublisher().stream().filter(publisher -> Objects.equals(publisher.getId(), edge.getSourceId()))
                .findFirst().orElseThrow();
            return Optional.of(RemoveCommand.create(editingDomain, pubs, PtnetPackage.Literals.PUBLISHERS__PUBLISHER, sub));
         }

      } else if (PNMLModelTypes.SERVICE_CHANNEL_LINK.equals(gElement.getType())){

      } else if (PNMLModelTypes.SERVICE_SERVER_LINK.equals(gElement.getType())){

      }
      return Optional.empty();
   }

}
