package org.eclipse.glsp.example.javaemf.handler;

import jakarta.inject.Inject;
import org.eclipse.emf.common.command.Command;
import org.eclipse.emf.common.command.CompoundCommand;
import org.eclipse.emf.edit.command.SetCommand;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.example.javaemf.validators.PnmlLabelEditValidator;
import org.eclipse.glsp.example.javaemf.model.ModelAccessUtil;
import org.eclipse.glsp.server.emf.EMFOperationHandler;
import org.eclipse.glsp.server.features.directediting.ApplyLabelEditOperation;
import org.eclipse.glsp.server.features.directediting.ValidationStatus;
import org.eclipse.glsp.server.types.Severity;
import org.pnml.emf.ptnet.*;

import java.util.Objects;
import java.util.Optional;

public class LabelEditOperationHandler extends EMFOperationHandler<ApplyLabelEditOperation> {
    @Inject
    protected PnmlLabelEditValidator labelValidator;

    @Inject
    protected ModelAccessUtil modelAccessUtil;


    @Override
    public Optional<Command> createCommand(ApplyLabelEditOperation operation) {
        //TODO emit markers instead of errors
        String labelId = operation.getLabelId();
        String labelText = operation.getText();


        if(labelId.endsWith("_label")) {
            String labeledElementId = labelId.replace("_label", "");
            return editLabel(labelText, labeledElementId);
        }
        if(labelId.endsWith("_arcWeight")){
            String labeledElementId = labelId.replace("_arcWeight", "");
            return editArcWeight(labelText, labeledElementId);
        }
        if(labelId.endsWith("_topicLinkLimitLabel")){
            labelId = labelId.replace("_topicLinkLimitLabel", "");
            if(labelId.endsWith("_subscriber")){
                String labeledElementId = labelId.replace("_subscriber", "");
                return editSubscriberLimit(labelText, labeledElementId);
            }
            if(labelId.endsWith("_publisher")){
                String labeledElementId = labelId.replace("_publisher", "");
                return editPublisherLimit(labelText, labeledElementId);
            }
        }
        if(labelId.endsWith("_channelLabel")){
            String labeledElementId = labelId.replace("_channelLabel", "");
            return editChannelId(labelText, labeledElementId);
        }

        return Optional.empty();
    }

    protected Optional<Command> editArcWeight(String labelText, String labeledElementId){
        Arc labeledElement = modelState.getIndex().getEObject(labeledElementId, Arc.class).orElseThrow(
            () -> new IllegalArgumentException("Label has no corresponding semantic model element"));

        ValidationStatus status = labelValidator.validateArcWeight(labelText);
        if(status.getSeverity().equals(Severity.ERROR)){
            throw new IllegalArgumentException(status.getMessage());
        }

        PTArcAnnotation annotation = PtnetFactory.eINSTANCE.createPTArcAnnotation();
        annotation.setText(Integer.parseInt(labelText));

        Command weightCommand = SetCommand.create(modelState.getEditingDomain(), labeledElement,
            PtnetPackage.Literals.EDGE__INSCRIPTION, annotation);

        return Optional.of(weightCommand);
    }

    protected Optional<Command> editLabel(String labelText, String labeledElementId){

        PnObject labeledElement = modelState.getIndex().getEObject(labeledElementId, PnObject.class).orElseThrow(
            () -> new IllegalArgumentException("Label has no corresponding semantic model element"));


        ValidationStatus status = labelValidator.validateName(labelText);
        if(status.getSeverity().equals(Severity.ERROR)){
            throw new IllegalArgumentException(status.getMessage());
        }

        Name name = labeledElement.getName();

        CompoundCommand command = new CompoundCommand();

        command.append(SetCommand.create(modelState.getEditingDomain(), name,
            PtnetPackage.Literals.NAME__TEXT, labelText));

        if(labeledElement instanceof Transition transition){
            if(modelAccessUtil.isTopicTransition(transition)) {
                ToolInfoTransition toolSpecifics = modelAccessUtil.getDiNeROSTransitionToolSpecific(transition, PNMLModelTypes.TOPIC_TRANSITION).orElseThrow();
                command.append(SetCommand.create(modelState.getEditingDomain(), toolSpecifics,
                    PtnetPackage.Literals.TOOL_INFO_TRANSITION__TOPIC_NAME, labelText));
            } else if(modelAccessUtil.isServiceTransition(transition)){
                ToolInfoTransition toolSpecifics = modelAccessUtil.getDiNeROSTransitionToolSpecific(transition, PNMLModelTypes.SERVICE_TRANSITION).orElseThrow();
                command.append(SetCommand.create(modelState.getEditingDomain(), toolSpecifics,
                    PtnetPackage.Literals.TOOL_INFO_TRANSITION__SERVICE_NAME, labelText));
            }
        }
        if(labeledElement instanceof Page page){
            if(modelAccessUtil.isServerPage(page)){
                ToolInfoPage toolSpecifics = modelAccessUtil.getDiNeROSPageToolSpecific(page).orElseThrow();
                command.append(SetCommand.create(modelState.getEditingDomain(), toolSpecifics,
                    PtnetPackage.Literals.TOOL_INFO_PAGE__SERVICE_NAME, labelText));
            }
        }




        return Optional.of(command);
    }

    protected Optional<Command> editSubscriberLimit(String labelText, String labelId){

        String[] idParts = labelId.split("_");
        String placeId = idParts[1];
        String labeledElementId = idParts[0];

        Transition labeledElement = modelState.getIndex().getEObject(labeledElementId, Transition.class).orElseThrow(
            () -> new IllegalArgumentException("Label has no corresponding semantic model element"));

        if(!modelAccessUtil.isTopicTransition(labeledElement)){
            throw new IllegalArgumentException("Transition is not a Topic");
        }

        ToolInfoTransition toolSpecifics = modelAccessUtil.getDiNeROSTransitionToolSpecific(labeledElement, PNMLModelTypes.TOPIC_TRANSITION).orElseThrow();

        ValidationStatus status = labelValidator.validateArcWeight(labelText);
        if(status.getSeverity().equals(Severity.ERROR)){
            throw new IllegalArgumentException(status.getMessage());
        }

        Subscriber subscriber = toolSpecifics.getSubscribers().getSubscriber().stream().filter(sub ->
            Objects.equals(sub.getId(), placeId)
        ).findFirst().orElseThrow(() -> new IllegalArgumentException("No subscriber with this id"));


        Command limitCommand = SetCommand.create(modelState.getEditingDomain(), subscriber,
            PtnetPackage.Literals.SUBSCRIBER__LIMIT, Integer.parseInt(labelText));

        return Optional.of(limitCommand);
    }

    protected Optional<Command> editPublisherLimit(String labelText, String labelId){
        String[] idParts = labelId.split("_");
        String placeId = idParts[1];
        String labeledElementId = idParts[0];

        Transition labeledElement = modelState.getIndex().getEObject(labeledElementId, Transition.class).orElseThrow(
            () -> new IllegalArgumentException("Label has no corresponding semantic model element"));

        if(!modelAccessUtil.isTopicTransition(labeledElement)){
            throw new IllegalArgumentException("Transition is not a Topic");
        }

        ToolInfoTransition toolSpecifics = modelAccessUtil.getDiNeROSTransitionToolSpecific(labeledElement, PNMLModelTypes.TOPIC_TRANSITION).orElseThrow();

        ValidationStatus status = labelValidator.validateArcWeight(labelText);
        if(status.getSeverity().equals(Severity.ERROR)){
            throw new IllegalArgumentException(status.getMessage());
        }

        Publisher publisher = toolSpecifics.getPublishers().getPublisher().stream().filter(pub ->
            Objects.equals(pub.getId(), placeId)
        ).findFirst().orElseThrow(() -> new IllegalArgumentException("No publisher with this id"));


        Command limitCommand = SetCommand.create(modelState.getEditingDomain(), publisher,
            PtnetPackage.Literals.PUBLISHER__LIMIT, Integer.parseInt(labelText));

        return Optional.of(limitCommand);
    }

    protected Optional<Command> editChannelId(String labelText, String labeledElementId){

        Channel labeledElement = modelState.getIndex().getEObject(labeledElementId, Channel.class).orElseThrow(
            () -> new IllegalArgumentException("Label has no corresponding semantic model element"));


        ValidationStatus status = labelValidator.validateName(labelText);
        if(status.getSeverity().equals(Severity.ERROR)){
            throw new IllegalArgumentException(status.getMessage());
        }

        Command cidCommand =  SetCommand.create(modelState.getEditingDomain(), labeledElement,
            PtnetPackage.Literals.CHANNEL__CID, labelText);


        return Optional.of(cidCommand);
    }
}
