package org.eclipse.glsp.example.javaemf.handler.NodeCreation;

import com.google.inject.Inject;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.example.javaemf.model.ExtendedModelState;
import org.eclipse.glsp.graph.*;
import org.eclipse.glsp.graph.util.GraphUtil;
import org.eclipse.glsp.server.emf.EMFCreateOperationHandler;
import org.eclipse.glsp.server.emf.EMFIdGenerator;
import org.eclipse.glsp.server.operations.CreateNodeOperation;
import org.eclipse.glsp.server.operations.CreateOperation;
import org.eclipse.glsp.server.utils.LayoutUtil;
import org.pnml.emf.ptnet.*;

import java.util.Optional;
import java.util.UUID;
import java.util.function.Function;

public abstract class AbstractCreateNodeHandler<T extends CreateOperation, ModelElement extends PnObject> extends EMFCreateOperationHandler<T> {
    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected EMFIdGenerator idGenerator;

    public AbstractCreateNodeHandler(final String... elementTypeIds) {
        super(elementTypeIds);
    }

    protected Optional<GPoint> getRelativeLocation(final CreateNodeOperation operation) {
        GModelElement container = modelState.getIndex().get(operation.getContainerId()).orElseGet(modelState::getRoot);

        Optional<GPoint> absoluteLocation = operation.getLocation();
        Optional<GPoint> relativeLocation = absoluteLocation.map(location-> LayoutUtil.getRelativeLocation(location, container));

        if(PNMLModelTypes.GRAPH.equals(container.getType())){
            // element was placed directly in the (top level of the) graph; no need for page adjustments
            return relativeLocation;
        }

        // slightly improves placement of nodes by removing the offset that is created by the compartment for the name of the page
        // as well as an apparent default padding
        // currently these offsets are: x:5, y:18
        try {
            double xOffsetCompartment = ((GCompartment) (container.getChildren().get(1))).getPosition().getX();
            double yOffsetCompartment = ((GCompartment) (container.getChildren().get(1))).getPosition().getY();

            relativeLocation.ifPresent(
                location -> {
                    location.setX(relativeLocation.get().getX() - xOffsetCompartment);
                    location.setY(relativeLocation.get().getY() - yOffsetCompartment);
                });
        } catch (Exception e){
            throw new RuntimeException("A problem occurred while accessing the structure of a page node. Perhaps the structure was changed in an unexpected way.");
        }

        return relativeLocation;
    }

    protected Page getContainerPage(final CreateNodeOperation operation) {
        return (Page) modelState.getIndex().getEObject(operation.getContainerId()).orElseThrow();
    }

    protected ModelElement initializeElement(ModelElement element) {
        setElementId(element);
        setInitialName(element);
        return element;
    }

    protected void setElementId(ModelElement element){
        element.setId(UUID.randomUUID().toString());
    }


    protected void setInitialName(final ModelElement modelElement) {
        setInitialName(modelElement, modelElement.eClass().getName());
    }

    protected void setInitialName(final ModelElement modelElement, String name) {
        Function<Integer, String> nameProvider = i -> "New" + name + i;
        int elementCounter = modelState.getIndex().getCounter(GraphPackage.Literals.GNODE, nameProvider);
        Name newName = PtnetFactory.eINSTANCE.createName();
        newName.setText(nameProvider.apply(elementCounter));
        modelElement.setName(newName);
    }



    protected NodeGraphics createShape(final String elementId, final Optional<GPoint> relativeLocation, final double width, final double height) {
        return createShape(elementId, relativeLocation, GraphUtil.dimension(width, height));
    }

    protected NodeGraphics createShape(final String elementId, final Optional<GPoint> relativeLocation, final GDimension size) {
        Position position = PtnetFactory.eINSTANCE.createPosition();
        position.setX(relativeLocation.map(GPoint::getX).orElse(0.0).intValue());
        position.setY(relativeLocation.map(GPoint::getY).orElse(0.0).intValue());

        Dimension dimension = PtnetFactory.eINSTANCE.createDimension();
        dimension.setX((int) size.getWidth());
        dimension.setY((int) size.getHeight());

        NodeGraphics newShape = PtnetFactory.eINSTANCE.createNodeGraphics();
        newShape.setPosition(position);
        newShape.setDimension(dimension);

        return newShape;
    }

    protected void setDiNeROSToolSpecifics(ToolInfoNode toolinfo, String pageNodeId){
        toolinfo.setTool(PNMLModelTypes.DINEROS_TOOL_NAME);
        toolinfo.setVersion(PNMLModelTypes.DINEROS_VERSION);
        toolinfo.setNode(pageNodeId);
        toolinfo.setSubnet("");
    }
}
