package org.eclipse.glsp.example.javaemf.model;

import jakarta.inject.Inject;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.example.javaemf.validators.InputSignalValidator;
import org.eclipse.glsp.graph.*;
import org.eclipse.glsp.graph.builder.impl.*;
import org.eclipse.glsp.graph.util.GConstants;
import org.eclipse.glsp.server.emf.EMFIdGenerator;
import org.pnml.emf.ptnet.*;

import java.util.*;

public class DiNeROSModelFactory {

    @Inject
    protected ExtendedModelState modelState;

    @Inject
    protected ModelAccessUtil modelAccessUtil;

    @Inject
    protected EMFIdGenerator idGenerator;

    @Inject
    protected InputSignalValidator inputSignalValidator;

    public List<GModelElement> createTopic(final Transition transition) {
        if (!modelAccessUtil.isTopicTransition(transition)){
            throw new IllegalArgumentException("Transition must be a topic transition.");
        }

        List<GModelElement> gElements = new ArrayList<>();

        GNodeBuilder topicNodeBuilder = new GNodeBuilder(PNMLModelTypes.TOPIC_TRANSITION)
            .id(idGenerator.getOrCreateId(transition))
            .addCssClass("topic-node")
            .add(new GLabelBuilder(DefaultTypes.LABEL)
                .text(transition.getName().getText())
                .id(transition.getId() + "_label")
                .build())
            .layout(GConstants.Layout.VBOX)
            .position(modelAccessUtil.getPosition(transition))
            .size(PNMLModelTypes.DefaultSizes.TopicWidth, PNMLModelTypes.DefaultSizes.TopicHeight)
            .addLayoutOptions(Map.of(
                GLayoutOptions.KEY_PREF_WIDTH, PNMLModelTypes.DefaultSizes.TopicWidth,
                GLayoutOptions.KEY_PREF_HEIGHT, PNMLModelTypes.DefaultSizes.TopicHeight,
                GLayoutOptions.KEY_PADDING_TOP, 5))
            .addArguments(GArguments.cornerRadius(20))
            ;

        gElements.add(topicNodeBuilder.build());

        ToolInfoTransition toolSpecifics = modelAccessUtil
            .getDiNeROSTransitionToolSpecific(transition, PNMLModelTypes.TOPIC_TRANSITION).orElseThrow();

        // iterates over all publishers and subscribers and creates corresponding links to them
        Optional.ofNullable(toolSpecifics.getPublishers()).map(Publishers::getPublisher).ifPresent(publishers ->
            publishers.forEach(publisher -> {
                if(modelState.getIndex().getEObject(publisher.getId()).isPresent()){
                    gElements.add(createTopicLink("publisher", publisher.getLimit(), publisher.getId(), transition.getId()));
                }
        }));
        Optional.ofNullable(toolSpecifics.getSubscribers()).map(Subscribers::getSubscriber).ifPresent(subscribers ->
            subscribers.forEach(subscriber -> {
                if(modelState.getIndex().getEObject(subscriber.getId()).isPresent()){
                    gElements.add(createTopicLink("subscriber", subscriber.getLimit(), transition.getId(), subscriber.getId()));
                }
        }));


        return gElements;

    }

    protected GEdge createTopicLink (String type, int limit, String sourceId, String targetId){
        String topicLimitId;
        if(type.equals("subscriber")){
            topicLimitId = sourceId + "_" + targetId + "_" + type + "_" + PNMLModelTypes.TOPIC_LINK_LIMIT_LABEL;
        } else {
            topicLimitId = targetId + "_" + sourceId + "_" + type + "_" + PNMLModelTypes.TOPIC_LINK_LIMIT_LABEL;

        }

        GLabelBuilder textLabelBuilder =  new GLabelBuilder(PNMLModelTypes.EDGE_TEXT_LABEL)
            .text(type)
            .id(sourceId + "_" + targetId + "_topicLinkLabel")
            .edgePlacement(new GEdgePlacementBuilder()
                .position(0.5)
                .side(GConstants.EdgeSide.BOTTOM)
                .rotate(true)
                .build()
            );

        GLabelBuilder limitLabelBuilder = new GLabelBuilder(PNMLModelTypes.TOPIC_LINK_LIMIT_LABEL)
            .text(String.valueOf(limit))
            .id(topicLimitId)
            .edgePlacement(new GEdgePlacementBuilder()
                .position(0.5)
                .side(GConstants.EdgeSide.TOP)
                .rotate(true)
                .build()
            );

        GEdgeBuilder topicLinkBuilder = new GEdgeBuilder(PNMLModelTypes.TOPIC_LINK)
            .id(sourceId + "_" + targetId + "_topicLink")
            .addCssClass("topic-link")
            .sourceId(sourceId)
            .targetId(targetId)
            .add(textLabelBuilder.build())
            .add(limitLabelBuilder.build());
        return topicLinkBuilder.build();
    }


    public List<GModelElement> createService(Transition transition){
        if(!modelAccessUtil.isServiceTransition(transition)){
            throw new IllegalArgumentException("Transition must be a service transition.");
        }

        List<GModelElement> gElements = new ArrayList<>();

        GNodeBuilder serviceNodeBuilder = new GNodeBuilder(PNMLModelTypes.SERVICE_TRANSITION)
            .id(idGenerator.getOrCreateId(transition))
            .addCssClass("service-node")
            .add(new GLabelBuilder(DefaultTypes.LABEL)
                .text(modelAccessUtil.getName(transition))
                .id(transition.getId() + "_label")
                .build())
            .layout(GConstants.Layout.VBOX)
            .position(modelAccessUtil.getPosition(transition))
            .addLayoutOptions(Map.of(
                GLayoutOptions.KEY_PREF_WIDTH, PNMLModelTypes.DefaultSizes.ServiceWidth,
                GLayoutOptions.KEY_PREF_HEIGHT, PNMLModelTypes.DefaultSizes.ServiceHeight,
                GLayoutOptions.KEY_PADDING_TOP, 5
                ))
            .addArguments(GArguments.cornerRadius(20));


        GCompartmentBuilder serviceCompartmentBuilder = new GCompartmentBuilder(PNMLModelTypes.PAGE)
            .id(idGenerator.getOrCreateId(transition) + "_comp")
            .layout(GConstants.Layout.FREEFORM)
            .size(PNMLModelTypes.DefaultSizes.ServiceWidth, PNMLModelTypes.DefaultSizes.ServiceHeight)
            .layoutOptions(Map.of(
                GLayoutOptions.KEY_PREF_WIDTH, PNMLModelTypes.DefaultSizes.ServiceWidth,
                GLayoutOptions.KEY_PREF_HEIGHT, PNMLModelTypes.DefaultSizes.ServiceHeight,
                GLayoutOptions.KEY_PADDING_TOP, 5,
                GLayoutOptions.KEY_V_GAP, 30,
                "vGrab", true,
                "hGrab", true
            ))
            .layoutOptions(new GLayoutOptions()
                .hAlign(GConstants.HAlign.LEFT)

            );


        // edges have to be added after the service Node, so they are displayed on top
        List<GEdge> gEdges = new ArrayList<>();

        ToolInfoTransition toolSpecifics = modelAccessUtil
            .getDiNeROSTransitionToolSpecific(transition, PNMLModelTypes.SERVICE_TRANSITION).orElseThrow();

        Optional.ofNullable(toolSpecifics.getChannels()).map(Channels::getChannel).ifPresent(channels -> {
            // pseudo final variable for lambda (equal to int spacing = 0)
            // required to calculate position of channel nodes in service node
            final int[] spacing = {0};
            channels.forEach(channel -> {
                serviceCompartmentBuilder.add(createChannelNode(channel, spacing[0]));
                spacing[0]++;
                gEdges.addAll(createChannelLink(channel));
        });});


        serviceNodeBuilder.add(serviceCompartmentBuilder.build());
        gElements.add(serviceNodeBuilder.build());
        gElements.addAll(gEdges);
        gElements.addAll(createServiceServerLink(toolSpecifics));

        return gElements;
    }

    protected GNode createChannelNode (Channel channel, int spacing){
        GNodeBuilder serviceNodeBuilder = new GNodeBuilder(PNMLModelTypes.SERVICE_CHANNEL)
            .id(channel.getCid())
            .addCssClass("service-channel")
            .size(PNMLModelTypes.DefaultSizes.ServiceChannelWidth, PNMLModelTypes.DefaultSizes.ServiceChannelHeight)
            .addLayoutOptions(Map.of(
                    GLayoutOptions.KEY_PREF_WIDTH, PNMLModelTypes.DefaultSizes.ServiceChannelWidth,
                    GLayoutOptions.KEY_PREF_HEIGHT, PNMLModelTypes.DefaultSizes.ServiceChannelHeight))
            .position(0,20 + spacing * (20 + PNMLModelTypes.DefaultSizes.ServiceChannelHeight))
            .add(new GLabelBuilder(DefaultTypes.LABEL)
                .text(channel.getCid())
                .id(channel.getCid() + "_channelLabel")
                .build())
            .layout(GConstants.Layout.VBOX)
            .addArguments(GArguments.cornerRadius(5));

        return serviceNodeBuilder.build();
    }

    protected List<GEdge> createChannelLink(Channel channel){
        List<GEdge> gEdges = new ArrayList<>();
        String request = channel.getRequest();
        if(request != null && !request.isEmpty() && modelState.getIndex().getEObject(request).isPresent()){
            GEdgeBuilder requestLinkBuilder = new GEdgeBuilder(PNMLModelTypes.SERVICE_CHANNEL_LINK);
            requestLinkBuilder
                .id(channel.getCid() + "_requestLink")
                .addCssClass("service-channel-link")
                .sourceId(request)
                .targetId(channel.getCid())
                .add(createEdgeLabel("request", channel.getCid()));
            gEdges.add(requestLinkBuilder.build());
        }
        String response = channel.getResponse();
        if(response != null && !response.isEmpty() && modelState.getIndex().getEObject(response).isPresent()){
            GEdgeBuilder responseLinkBuilder = new GEdgeBuilder(PNMLModelTypes.SERVICE_CHANNEL_LINK);
            responseLinkBuilder
                .id(channel.getCid() + "_responseLink")
                .addCssClass("service-channel-link")
                .sourceId(channel.getCid())
                .targetId(response)
                .add(createEdgeLabel("response", channel.getCid()));
            gEdges.add(responseLinkBuilder.build());
        }

        return  gEdges;
    }

    protected List<GEdge> createServiceServerLink(ToolInfoTransition toolSpecifics){
        List<GEdge> gEdges = new ArrayList<>();
        String request = toolSpecifics.getServerInput();
        if(request != null && !request.isEmpty() && modelState.getIndex().getEObject(request).isPresent()){
            GEdgeBuilder requestLinkBuilder = new GEdgeBuilder(PNMLModelTypes.SERVICE_SERVER_LINK);
            requestLinkBuilder
                .id(request + "_" + toolSpecifics.getTransition().getId() + "_requestLink")
                .addCssClass("service-server-link")
                .sourceId(request)
                .targetId(toolSpecifics.getTransition().getId())
                .add(createEdgeLabel("request", request + "_" + toolSpecifics.getTransition().getId()));
            gEdges.add(requestLinkBuilder.build());
        }
        String response = toolSpecifics.getServerOutput();
        if(response != null && !response.isEmpty() && modelState.getIndex().getEObject(response).isPresent()){
            GEdgeBuilder responseLinkBuilder = new GEdgeBuilder(PNMLModelTypes.SERVICE_CHANNEL_LINK);
            responseLinkBuilder
                .id(toolSpecifics.getTransition().getId() + "_" + request + "_responseLink")
                .addCssClass("service-server-link")
                .sourceId(toolSpecifics.getTransition().getId())
                .targetId(response)
                .add(createEdgeLabel("response", toolSpecifics.getTransition().getId() + "_" + response));
            gEdges.add(responseLinkBuilder.build());
        }

        return  gEdges;
    }

    public GLabel createEdgeLabel (String text, String nodeId){
        return new GLabelBuilder(PNMLModelTypes.EDGE_TEXT_LABEL)
            .id(nodeId + "_" + text + "_linkLabel")
            .text(text)
            .edgePlacement(new GEdgePlacementBuilder()
                .position(0.5)
                .side(GConstants.EdgeSide.BOTTOM)
                .rotate(true)
                .build()
            ).build();
    }

    public GNode createInputSignalClauseNode(Transition transition){
        String signalClause = modelAccessUtil.getInputSignalClause(transition);

        GPoint parentPosition = modelAccessUtil.getPosition(transition);

        GLabelBuilder labelBuilder = new GLabelBuilder(PNMLModelTypes.SIGNAL_CLAUSE_LABEL)
            .text(signalClause)
            .id(transition.getId() + "_signalClause");
        if(signalClause.isEmpty() || inputSignalValidator.evaluateCNF(signalClause)){
            labelBuilder.addCssClass("signal-clause-true");
        } else {
            labelBuilder.addCssClass("signal-clause-false");
        }

        GNodeBuilder labelNodeBuilder = new GNodeBuilder(PNMLModelTypes.LABEL_NODE)
            .id(transition.getId() + "_signalClauseLabelNode")
            .layout(GConstants.Layout.VBOX)
            .size(200, 20)
            .addCssClass("label-node")
            .addLayoutOptions(Map.of(
                GLayoutOptions.KEY_PREF_WIDTH, 200,
                GLayoutOptions.KEY_PREF_HEIGHT, 20,
                GLayoutOptions.KEY_H_ALIGN, GConstants.HAlign.CENTER))
            .position(parentPosition.getX()-((double) (200 - PNMLModelTypes.DefaultSizes.TransitionWidth) /2),
                parentPosition.getY() - 20)
            .add(labelBuilder.build())
            ;

        return labelNodeBuilder.build();
    }


}
