/********************************************************************************
 * Copyright (c) 2022 EclipseSource and others.
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v. 2.0 which is available at
 * https://www.eclipse.org/legal/epl-2.0.
 *
 * This Source Code may also be made available under the following Secondary
 * Licenses when the conditions for such availability set forth in the Eclipse
 * Public License v. 2.0 are satisfied:
 * -- GNU General Public License, version 2 with the GNU Classpath Exception
 * which is available at https://www.gnu.org/software/classpath/license.html
 * -- MIT License which is available at https://opensource.org/license/mit.
 *
 * SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0 OR MIT
 ********************************************************************************/
package org.eclipse.glsp.example.javaemf.model;

import java.util.*;

import com.google.gson.Gson;
import jakarta.inject.Inject;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.glsp.example.javaemf.MQTT.MQTTClient;
import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.graph.*;
import org.eclipse.glsp.graph.builder.impl.*;
import org.eclipse.glsp.graph.util.GConstants;
import org.eclipse.glsp.server.emf.model.notation.Diagram;
import org.eclipse.glsp.server.emf.notation.EMFNotationGModelFactory;
import org.pnml.emf.ptnet.*;


public class TaskListGModelFactory extends EMFNotationGModelFactory {

   @Inject
   protected ExtendedModelState modelState;

   @Inject
   protected ModelAccessUtil modelAccessUtil;

   @Inject
   protected DiNeROSModelFactory diNeROSModelFactory;

   @Inject
   protected MQTTClient mqttClient;

   protected Gson gson = new Gson();

   @Override
   protected void fillRootElement(final EObject semanticModel, final Diagram notationModel, final GModelRoot newRoot) {
      pnml netContainer = pnml.class.cast(semanticModel);
      PetriNet petriNet = netContainer.getNet().stream().findFirst().orElseThrow();
      GGraph graph = GGraph.class.cast(newRoot);
      //iterates over all top level pages
      petriNet.getPage().stream().forEachOrdered(page -> {
         graph.getChildren().add(createPage(page));
      });

      addSignals(graph);
   }

   @Override
   protected void fillRootElement(final GModelRoot newRoot) {
      Diagram notationModel = modelState.getNotationModel();
      EObject semanticModel = modelState.getSemanticModel();
      //modelState.getIndex().indexAll(notationModel, semanticModel);
      modelState.getIndex().indexSemanticModel(semanticModel);
      fillRootElement(semanticModel, notationModel, newRoot);
   }


   protected List<GModelElement> createPlace(final Place place) {
      List<GModelElement> gNodes = new ArrayList<>();

      int marking = modelAccessUtil.getTokensInPlace(place);

      GNodeBuilder placeNodeBuilder = new GNodeBuilder(PNMLModelTypes.PLACE)
          .id(idGenerator.getOrCreateId(place))
          .addCssClass("place")
          .add(new GLabelBuilder(PNMLModelTypes.MARKING_LABEL)
              .text(Integer.toString(marking))
              .id(place.getId() + "_marking")
              .build())
          .layout(GConstants.Layout.VBOX)
          .layoutOptions(new GLayoutOptions()
              .paddingTop((PNMLModelTypes.DefaultSizes.PlaceHeight-10)/2)
              .vGap(30)
              .hAlign(GConstants.HAlign.CENTER)
          )
          .addArgument("SimulationMarking", marking)
          .addArgument("InitialMarking", modelAccessUtil.getInitialMarking(place))
          .addArgument("BalloonTokens", gson.toJson(modelAccessUtil.getBalloonTokens(place)))
          ;

      applyNodeShape(place,PNMLModelTypes.DefaultSizes.PlaceWidth, PNMLModelTypes.DefaultSizes.PlaceHeight, placeNodeBuilder);

      gNodes.add(placeNodeBuilder.build());
      gNodes.add(createLabelNode(place, PNMLModelTypes.DefaultSizes.PlaceWidth, PNMLModelTypes.DefaultSizes.PlaceHeight));
      return gNodes;
   }


   protected GNodeBuilder applyNodeShape(Node node, double width, double height, GNodeBuilder builder) {
      GPoint position = modelAccessUtil.getPosition(node);

      builder
          .position(position.getX(), position.getY())
          .size(width, height)
          .addLayoutOptions(Map.of(
              GLayoutOptions.KEY_PREF_WIDTH, width,
              GLayoutOptions.KEY_PREF_HEIGHT, height));

      return builder;
   }

   protected GNodeBuilder applyPageShape(Page page, GNodeBuilder builder) {
      GPoint position = modelAccessUtil.getPosition(page);

      double width = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getDimension).map(Coordinate::getX).orElse(200);
      double height = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getDimension).map(Coordinate::getY).orElse(200);

      builder
          .position(position.getX(), position.getY())
          .size(width, height)
          .addLayoutOptions(Map.of(
              GLayoutOptions.KEY_PREF_WIDTH, width,
              GLayoutOptions.KEY_PREF_HEIGHT, height));

      return builder;
   }


   protected List<GModelElement> createTransition(final Transition transition) {
      if(modelAccessUtil.isTopicTransition(transition)){
         return diNeROSModelFactory.createTopic(transition);
      }
      if(modelAccessUtil.isServiceTransition(transition)){
         return diNeROSModelFactory.createService(transition);
      }


      List<GModelElement> gElements = new ArrayList<>();

      GNodeBuilder transitionNodeBuilder = new GNodeBuilder(PNMLModelTypes.TRANSITION)
          .id(idGenerator.getOrCreateId(transition))
          .addCssClass("transition")
          .layout(GConstants.Layout.HBOX, Map.of(GLayoutOptions.KEY_PADDING_LEFT, 5));

      applyNodeShape(transition, PNMLModelTypes.DefaultSizes.TransitionWidth, PNMLModelTypes.DefaultSizes.TransitionHeight, transitionNodeBuilder);

      gElements.add(transitionNodeBuilder.build());
      gElements.add(createLabelNode(transition, PNMLModelTypes.DefaultSizes.TransitionWidth, PNMLModelTypes.DefaultSizes.TransitionHeight));
      gElements.add(diNeROSModelFactory.createInputSignalClauseNode(transition));

      return gElements;
   }


   protected GEdge createArc(final Arc arc) {
      GEdgeBuilder arcBuilder = new GEdgeBuilder(PNMLModelTypes.ARC)
          .id(idGenerator.getOrCreateId(arc))
          .addCssClass("arc")
          .sourceId(arc.getSource().getId())
          .targetId(arc.getTarget().getId())
          .add(new GLabelBuilder(PNMLModelTypes.ARC_WEIGHT_LABEL)
              .text(String.valueOf(modelAccessUtil.getArcWeight(arc)))
              .id(idGenerator.getOrCreateId(arc) + "_arcWeight")
              .edgePlacement(new GEdgePlacementBuilder()
                  .rotate(true)
                  .position(0.5)
                  .side(GConstants.EdgeSide.BOTTOM) // if rotate is set to true, BOTTOM actually renders the label on top of the edge
                  .build()
              )
              .build())
          ;

      if(modelAccessUtil.isInhibitorArc(arc)){
         arcBuilder
             .type(PNMLModelTypes.INHIBITOR_ARC)
             .addCssClass("inhibitor-arc");
      }
      if(modelAccessUtil.isReadArc(arc)){
         arcBuilder
             .type(PNMLModelTypes.READ_ARC)
             .addCssClass("read-arc");
      }

      applyEdgeData(arc, arcBuilder);
      return arcBuilder.build();
   }


   protected GNode createPage(final Page page) {
      Map<String, Object> layoutOptions = new HashMap<>();
      layoutOptions.put(GLayoutOptions.KEY_H_ALIGN, GConstants.HAlign.LEFT);

      Map<String, Object> pageLayoutOptions = new HashMap<>();

      //double x = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getPosition).map(Coordinate::getX).orElse(0);
      //double y = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getPosition).map(Coordinate::getY).orElse(0);;
      //double width = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getDimension).map(Coordinate::getX).orElse(200);;
      //double height = Optional.ofNullable(page.getNodegraphics()).map(NodeGraphics::getDimension).map(Coordinate::getY).orElse(200);;


      GCompartmentBuilder pageBuilder = new GCompartmentBuilder(PNMLModelTypes.PAGE)
          .id(idGenerator.getOrCreateId(page) + "_comp")
          //.size(width, height)
          //.position(x, y)
          .addCssClass("page-comp")
          .layout(GConstants.Layout.FREEFORM)
          .layoutOptions(pageLayoutOptions);

      //recursively adds all nodes to the page and the nodes to its subpages
      populatePage(page, pageBuilder);

      GNodeBuilder pageNodeBuilder = new GNodeBuilder(PNMLModelTypes.PAGE_NODE)
          .id(idGenerator.getOrCreateId(page))
          .layout(GConstants.Layout.VBOX)
          .layoutOptions(layoutOptions)
          .addArguments(GArguments.cornerRadius(20))
          .addArgument("hidden", Optional.ofNullable( modelState.getHiddenPages().get(page.getId())).orElse(false));

      if(modelAccessUtil.isServerPage(page)){
         pageNodeBuilder
             .type(PNMLModelTypes.SERVER_PAGE)
             .addCssClass("server-page")
             .add(new GLabelBuilder(DefaultTypes.LABEL)
                 .text((modelAccessUtil.getName(page)))
                 .id(idGenerator.getOrCreateId(page) + "_label")
                 .build());
      } else 
      if(modelAccessUtil.isNodePageNode(page)){
        pageNodeBuilder
            .type(PNMLModelTypes.NODE_PAGE_NODE)
            .addCssClass("nodePage-node")
            .add(new GLabelBuilder(DefaultTypes.LABEL)
                .text((modelAccessUtil.getName(page)))
                .id(idGenerator.getOrCreateId(page) + "_label")
                .build());
      } else {
         pageNodeBuilder
             .type(PNMLModelTypes.PAGE_NODE)
             .addCssClass("page-node")
             .add(new GLabelBuilder(DefaultTypes.LABEL)
                 .text(modelAccessUtil.getName(page))
                 .id(idGenerator.getOrCreateId(page) + "_label")
                 .build());
      }

      pageNodeBuilder.add(pageBuilder.build());
      applyPageShape(page, pageNodeBuilder);
      return pageNodeBuilder.build();
      /**/
   }


   protected GCompartmentBuilder populatePage(Page page, GCompartmentBuilder pageBuilder) {
      page.getPlace().stream()
          .map(this::createPlace)
          .forEachOrdered(pageBuilder::addAll);
      page.getTransition().stream()
          .map(this::createTransition)
          .forEachOrdered(pageBuilder::addAll);

      page.getReferencePlace().stream()
          .map(this::createRefPlace)
          .forEachOrdered(pageBuilder::addAll);
      page.getReferenceTransition().stream()
          .map(this::createRefTransition)
          .forEachOrdered(pageBuilder::addAll);

      page.getArc().stream()
          .map(this::createArc)
          .forEachOrdered(pageBuilder::add);

      page.getPage().stream()
          .map(this::createPage)
          .forEachOrdered(pageBuilder::add);

      return pageBuilder;
   }


   protected List<GModelElement> createRefPlace(final RefPlace refPlace) {
      List<GModelElement> gElements = new ArrayList<>();
      int marking = 0;
      PlaceNode referencedPlace = refPlace.getRef();
      while (referencedPlace instanceof RefPlace){
         referencedPlace = ((RefPlace) referencedPlace).getRef();
      }
      if (referencedPlace instanceof Place place){
         marking = modelAccessUtil.getTokensInPlace(place);
      }

      GNodeBuilder refPlaceNodeBuilder = new GNodeBuilder(PNMLModelTypes.REF_PLACE)
          .id(idGenerator.getOrCreateId(refPlace))
          .addCssClass("ref-place")
          .add(new GLabelBuilder(PNMLModelTypes.MARKING_LABEL)
              .text(Integer.toString(marking))
              .id(refPlace.getId() + "_marking")
              .build())
          .layout(GConstants.Layout.VBOX)
          .layoutOptions(new GLayoutOptions()
              .paddingTop(30)
              .vGap(30)
              .hAlign(GConstants.HAlign.CENTER)
          )
          .addArgument("marking", marking);

      applyNodeShape(refPlace,PNMLModelTypes.DefaultSizes.PlaceWidth, PNMLModelTypes.DefaultSizes.PlaceHeight, refPlaceNodeBuilder);

      gElements.add(refPlaceNodeBuilder.build());
      gElements.add(createLabelNode(refPlace, PNMLModelTypes.DefaultSizes.PlaceWidth, PNMLModelTypes.DefaultSizes.PlaceHeight));
      if(refPlace.getRef() != null){
         gElements.add(createRefLink(refPlace.getId(), refPlace.getRef().getId()));
      }

      return gElements;
   }


   protected List<GModelElement> createRefTransition(final RefTransition refTransition) {
      List<GModelElement> gElements = new ArrayList<>();
      GNodeBuilder refTransitionNodeBuilder = new GNodeBuilder(PNMLModelTypes.REF_TRANSITION)
          .id(idGenerator.getOrCreateId(refTransition))
          .addCssClass("ref-transition")
          .layout(GConstants.Layout.HBOX, Map.of(GLayoutOptions.KEY_PADDING_LEFT, 5));

      applyNodeShape(refTransition, PNMLModelTypes.DefaultSizes.TransitionWidth, PNMLModelTypes.DefaultSizes.TransitionHeight, refTransitionNodeBuilder);

      gElements.add(refTransitionNodeBuilder.build());
      gElements.add(createLabelNode(refTransition, PNMLModelTypes.DefaultSizes.TransitionWidth, PNMLModelTypes.DefaultSizes.TransitionHeight));
      if(refTransition.getRef() != null){
         gElements.add(createRefLink(refTransition.getId(), refTransition.getRef().getId()));
      }

      return gElements;
   }


   protected GEdge createRefLink(final String sourceId, final String targetId){
      GEdgeBuilder refLinkBuilder = new GEdgeBuilder(PNMLModelTypes.REF_LINK)
          .id(sourceId + "_" +targetId + "_refLink")
          .addCssClass("ref-link")
          .sourceId(sourceId)
          .targetId(targetId);
      return refLinkBuilder.build();
   }


   protected GNode createLabelNode(Node parent, double width, double height){
      GPoint parentPosition = modelAccessUtil.getPosition(parent);

      GNodeBuilder labelNodeBuilder = new GNodeBuilder(PNMLModelTypes.LABEL_NODE)
          .id(parent.getId() + "_labelNode")
          .layout(GConstants.Layout.VBOX)
          .size(100, 20)
          .addCssClass("label-node")
          .addLayoutOptions(Map.of(
              GLayoutOptions.KEY_PREF_WIDTH, 100,
              GLayoutOptions.KEY_PREF_HEIGHT, 20,
              GLayoutOptions.KEY_H_ALIGN, GConstants.HAlign.CENTER))
          .position(parentPosition.getX()-((100-width)/2), parentPosition.getY()+height)
          .add(new GLabelBuilder(DefaultTypes.LABEL)
              .text(modelAccessUtil.getName(parent))
              .id(parent.getId() + "_label")
              .build())
          ;
      return labelNodeBuilder.build();
   }


   protected void addSignals(GModelRoot graph){
      pnml netContainer = modelState.getSemanticModel(pnml.class).orElseThrow();
      PetriNet petriNet = netContainer.getNet().stream().findFirst().orElseThrow();
      HashSet<String> signals = new HashSet<>();

      Optional<ToolInfoNet> toolSpecifics = modelAccessUtil.getDiNeROSNetToolSpecific(petriNet);
      if (toolSpecifics.isEmpty()){
         updateMQTTClientSignals(signals);
         return;
      }
      if(toolSpecifics.get().getInputsignals() != null){
         toolSpecifics.get().getInputsignals().getInputsignal().forEach(inputSignal -> {
            graph.getArgs().put(inputSignal.getInputsignalID(), inputSignal.isInitialvalue());
            signals.add(inputSignal.getInputsignalID());
         });
      }
      updateMQTTClientSignals(signals);
   }


   protected void updateMQTTClientSignals(HashSet<String> signals){
      String[] topics = new String[mqttClient.getSignalTopics().size()];
      mqttClient.getSignalTopics().toArray(topics);
      signals.forEach(signal -> {mqttClient.subscribe(signal);});
       for (String s : topics) {
           if (!signals.contains(s)) {
               mqttClient.unsubscribe(s);
           }
       }
   }

}
