package org.eclipse.glsp.example.javaemf.validators;

import org.eclipse.glsp.example.javaemf.PNMLModelTypes;
import org.eclipse.glsp.graph.GModelElement;
import org.eclipse.glsp.server.features.typehints.EdgeCreationChecker;

import java.util.List;
import java.util.Map;
import java.util.Objects;

import static java.util.Map.entry;

public class PNMLEdgeCreationChecker implements EdgeCreationChecker {
    Map<String, List<String>> validSourceElements = Map.ofEntries(
        entry(PNMLModelTypes.ARC, List.of(
            PNMLModelTypes.TRANSITION,
            PNMLModelTypes.REF_TRANSITION,
            PNMLModelTypes.PLACE,
            PNMLModelTypes.REF_PLACE)),
        entry(PNMLModelTypes.REFERENCE, List.of(
            PNMLModelTypes.REF_TRANSITION,
            PNMLModelTypes.REF_PLACE)),
        entry(PNMLModelTypes.TOPIC_LINK,
            List.of(PNMLModelTypes.PLACE, PNMLModelTypes.TOPIC_TRANSITION)),
        entry(PNMLModelTypes.SERVICE_SERVER_LINK,
            List.of(PNMLModelTypes.PLACE, PNMLModelTypes.SERVICE_TRANSITION)),
        entry(PNMLModelTypes.SERVICE_CHANNEL_LINK,
            List.of(PNMLModelTypes.PLACE, PNMLModelTypes.SERVICE_CHANNEL))
    );

    @Override
    public boolean isValidSource(final String edgeType, final GModelElement sourceElement) {
        return validSourceElements.containsKey(edgeType) && validSourceElements.get(edgeType).contains(sourceElement.getType());
    }

    @Override
    public boolean isValidTarget(final String edgeType, final GModelElement sourceElement, final GModelElement targetElement) {
        // checks for each "dynamic" edge type whether the source and target node are allowed to be connected
        switch (edgeType) {
            case PNMLModelTypes.ARC:
                if (!Objects.equals(sourceElement.getParent(), targetElement.getParent())){
                    // arcs can only connect nodes on the same page
                    return false;
                }
                if (PNMLModelTypes.TRANSITION.equals(sourceElement.getType())
                    || PNMLModelTypes.REF_TRANSITION.equals(sourceElement.getType())) {
                    return PNMLModelTypes.PLACE.equals(targetElement.getType())
                        || PNMLModelTypes.REF_PLACE.equals(targetElement.getType());
                }
                else if (PNMLModelTypes.PLACE.equals(sourceElement.getType())
                    || PNMLModelTypes.REF_PLACE.equals(sourceElement.getType())){
                    return PNMLModelTypes.TRANSITION.equals(targetElement.getType())
                        || PNMLModelTypes.REF_TRANSITION.equals(targetElement.getType());
                } else return false;
            case PNMLModelTypes.REFERENCE:
                if (PNMLModelTypes.REF_TRANSITION.equals(sourceElement.getType())) {
                    return PNMLModelTypes.TRANSITION.equals(targetElement.getType())
                        || PNMLModelTypes.REF_TRANSITION.equals(targetElement.getType());
                }
                else if (PNMLModelTypes.REF_PLACE.equals(sourceElement.getType())){
                    return PNMLModelTypes.PLACE.equals(targetElement.getType())
                        || PNMLModelTypes.REF_PLACE.equals(targetElement.getType());
                } else return false;
            case PNMLModelTypes.TOPIC_LINK:
                if (PNMLModelTypes.PLACE.equals(sourceElement.getType())){
                    return PNMLModelTypes.TOPIC_TRANSITION.equals(targetElement.getType());
                }
                else if (PNMLModelTypes.TOPIC_TRANSITION.equals(sourceElement.getType())){
                    return PNMLModelTypes.PLACE.equals(targetElement.getType());
                } else return false;
            case PNMLModelTypes.SERVICE_SERVER_LINK:
                if (PNMLModelTypes.PLACE.equals(sourceElement.getType())){
                    return PNMLModelTypes.SERVICE_TRANSITION.equals(targetElement.getType());
                }
                else if (PNMLModelTypes.SERVICE_TRANSITION.equals(sourceElement.getType())){
                    return PNMLModelTypes.PLACE.equals(targetElement.getType());
                } else return false;
            case PNMLModelTypes.SERVICE_CHANNEL_LINK:
                if (PNMLModelTypes.PLACE.equals(sourceElement.getType())){
                    return PNMLModelTypes.SERVICE_CHANNEL.equals(targetElement.getType());
                }
                else if (PNMLModelTypes.SERVICE_CHANNEL.equals(sourceElement.getType())){
                    return PNMLModelTypes.PLACE.equals(targetElement.getType());
                } else return false;
            default:
                return false;
        }
    }
}
